#!/usr/bin/env python

"""Showcase textual-kitty Renderables/Widgets."""

from argparse import ArgumentParser
from pathlib import Path
from typing import Type, cast

from PIL import Image as PILImage
from PIL import ImageOps

from textual_kitty.renderable._base import _Base

TEST_IMAGE = Path(__file__).parent / "gracehopper.jpg"


def demo_rich(Image: Type[_Base]) -> None:
    """Showcase textual-kitty's Rich Renderable for images."""
    from rich.console import Console
    from rich.style import Style
    from rich.table import Table

    console = Console()

    console.print("KittyImage from path:", style=Style(bold=True))
    console.print(Image(TEST_IMAGE))

    console.print("\nKittyImage from pillow image:", style=Style(bold=True))
    with PILImage.open(TEST_IMAGE) as image:
        console.print(Image(image))

    console.print("\nSame KittyImage instance rendered twice:", style=Style(bold=True))
    image = Image(TEST_IMAGE)
    console.print(image)
    console.print(image)

    console.print("\nKittyImage with width set to 10:", style=Style(bold=True))
    console.print(Image(TEST_IMAGE, width=10))

    console.print("\nKittyImage inside table:", style=Style(bold=True))
    table = Table("Name", "Birthday", "Picture")
    table.add_row("Grace Hopper", "December 9, 1906", Image(TEST_IMAGE))
    console.print(table)


def demo_textual(Renderable: Type[_Base]) -> None:
    """Showcase textual-kitty's Textual Widget for images."""
    from textual import on
    from textual.app import App, ComposeResult
    from textual.containers import Container, HorizontalScroll, VerticalScroll
    from textual.widgets import Button, Collapsible, Footer, Header

    from textual_kitty.widget import Image

    class DemoApp(App[None]):
        DEFAULT_CSS = """
        DemoApp {
            #content {
                layout: grid;
                grid-size: 5 2;
            }

            #content > * {
                border: round gray;
                align: center middle;
            }

            .small {
                width: 6;
            }

            .full-row {
                column-span: 5;
            }

            .border {
                border: round white;
            }

            .width-auto {
                width: auto;
            }

            .width-full {
                width: 100%;
            }
        }
        """

        def compose(self) -> ComposeResult:
            yield Header()
            with VerticalScroll(id="content"):
                with Container() as c:
                    c.border_title = "Grace from path"
                    yield Image(TEST_IMAGE, image_renderable_type=Renderable)
                with Container() as c:
                    c.border_title = "Grace from PIL image"
                    with PILImage.open(TEST_IMAGE) as image:
                        yield Image(image, image_renderable_type=Renderable)
                with Container() as c:
                    c.border_title = "Collapsible Grace"
                    with Collapsible(collapsed=False):
                        yield Image(TEST_IMAGE, image_renderable_type=Renderable)
                with Container() as c:
                    c.border_title = "Small Grace"
                    yield Image(TEST_IMAGE, image_renderable_type=Renderable, classes="small")
                with Container() as c:
                    c.border_title = "Flippable Grace"
                    with PILImage.open(TEST_IMAGE) as image:
                        yield Image(image, image_renderable_type=Renderable, id="flippable")
                    yield Button("Flip", classes="width-full")
                with HorizontalScroll(classes="full-row") as c:
                    c.border_title = "Many Graces (huge, loaded async)"
                    with PILImage.open(TEST_IMAGE) as image:
                        image = image.resize((image.width * 4, image.height * 4))
                    for _ in range(50):
                        yield Image(
                            image, image_renderable_type=Renderable, classes="border width-auto", load_async=True
                        )
            yield Footer()

        @on(Button.Pressed)
        def button_pressed(self, _event: Button.Pressed) -> None:
            widget = self.query_one("#flippable", Image)
            widget.image = ImageOps.flip(cast(PILImage.Image, widget.image))

    DemoApp().run()


def run() -> None:
    """Run the textual_kitty demo."""
    parser = ArgumentParser(description="Demo the capabilities of textual-kitty")
    parser.add_argument("mode", choices=["rich", "textual"], nargs="?", default="rich")
    parser.add_argument(
        "-p", "--protocol", choices=["auto", "tgp", "colored-fallback", "grayscale-fallback"], default="auto"
    )
    arguments = parser.parse_args()

    if arguments.protocol == "auto":
        from textual_kitty.renderable import Image
    elif arguments.protocol == "tgp":
        from textual_kitty.renderable.tgp import Image
    elif arguments.protocol == "colored-fallback":
        from textual_kitty.renderable.fallback.colored import Image
    elif arguments.protocol == "grayscale-fallback":
        from textual_kitty.renderable.fallback.grayscale import Image

    if arguments.mode == "rich":
        demo_rich(Image)
    elif arguments.mode == "textual":
        demo_textual(Image)


if __name__ == "__main__":
    run()



from array import array
from typing import IO
from unittest.mock import patch

from pytest import raises

from tests.data import TERM_SIZE, TERM_SIZE_NO_SCREEN
from textual_kitty.terminal import TerminalError, get_terminal_size_info


def test_get_terminal_size_info_on_tty_success() -> None:
    def ioctl(fd: IO[bytes], request: int, buf: array[int]) -> None:
        buf[0], buf[1], buf[2], buf[3], _, _ = TERM_SIZE

    with patch("sys.__stdout__.isatty", return_value=True):
        with patch("textual_kitty.terminal.ioctl", side_effect=ioctl):
            assert get_terminal_size_info() == TERM_SIZE


def test_get_terminal_size_info_on_tty_no_screen_size() -> None:
    def ioctl(fd: IO[bytes], request: int, buf: array[int]) -> None:
        buf[0], buf[1], buf[2], buf[3], _, _ = TERM_SIZE_NO_SCREEN

    with patch("sys.__stdout__.isatty", return_value=True):
        with patch("textual_kitty.terminal.ioctl", side_effect=ioctl):
            assert get_terminal_size_info() == TERM_SIZE_NO_SCREEN


def test_get_terminal_size_info_on_tty_failure() -> None:
    with patch("sys.__stdout__.isatty", return_value=True):
        with patch("textual_kitty.terminal.ioctl", side_effect=OSError()):
            with raises(TerminalError):
                get_terminal_size_info()


def test_get_terminal_size_info_stdout_closed() -> None:
    with patch("sys.__stdout__", None):
        with raises(TerminalError):
            get_terminal_size_info()


def test_get_terminal_size_info_stdout_not_a_tty() -> None:
    with patch("sys.__stdout__.isatty", return_value=False):
        term_size = get_terminal_size_info()

    assert term_size.screen_width == 0
    assert term_size.screen_height == 0



"""Functionality to interact with the terminal."""

import logging
import sys
import termios
from array import array
from fcntl import ioctl
from typing import NamedTuple

logger = logging.getLogger(__name__)


class TerminalError(Exception):
    """Error thrown on failing terminal operations."""

    pass


class TerminalSizeInformation(NamedTuple):
    """Size of several terminal features."""

    rows: int
    """Width of the terminal in cells."""
    columns: int
    """Height of the terminal in cells."""
    screen_width: int
    """Width of the terminal in pixels."""
    screen_height: int
    """Height of the terminal in pixels."""
    cell_width: int
    """Width of a terminal cell in pixels."""
    cell_height: int
    """Height of a terminal cell in pixels."""


def get_terminal_size_info() -> TerminalSizeInformation:
    """Get size information from the terminal.

    Returns:
        The size information

    """
    if not sys.__stdout__:
        raise TerminalError("stdout is closed")

    if sys.__stdout__.isatty():
        try:
            buf = array("H", [0, 0, 0, 0])
            ioctl(sys.__stdout__, termios.TIOCGWINSZ, buf)

            rows, columns, screen_width, screen_height = buf

            # If we can't get the cell size (not all terminals support this) we're just assuming some hopefully reasonable numbers.
            # Does this suck? Yes. Is there something we can do about it? Not really.
            if not screen_width or not screen_height:
                logger.warning("Terminal doesn't support getting pixel size; falling back to assumptions")

        except OSError as e:
            raise TerminalError("Unsupported terminal") from e

    else:
        logger.debug("Not connected to a terminal, assuming reasonable size information")
        rows = 24
        columns = 80
        screen_width = 0
        screen_height = 0

    cell_width = int(screen_width / columns) or 8
    cell_height = int(screen_height / rows) or 12

    return TerminalSizeInformation(
        rows=rows,
        columns=columns,
        screen_width=screen_width,
        screen_height=screen_height,
        cell_width=cell_width,
        cell_height=cell_height,
    )



"""Geometry data structures."""

from typing import NamedTuple


class Size(NamedTuple):
    """The dimensions (width and height) of a rectangular region."""

    width: int
    """The width."""
    height: int
    """The height."""



from unittest.mock import MagicMock, Mock, patch

import pytest
from PIL import Image as PILImage
from PIL import ImageOps
from textual.app import App, ComposeResult

from tests.data import SIZE_ZERO, TERM_SIZE, TEST_IMAGE
from textual_kitty.widget import Image


async def test_app() -> None:
    class TestApp(App[None]):
        CSS = """
        .auto {
            width: auto;
            height: auto;
        }
        """

        def compose(self) -> ComposeResult:
            yield Image(TEST_IMAGE)
            yield Image(TEST_IMAGE, classes="auto")
            yield Image()
            yield Image(classes="auto")

    app = TestApp()

    # Not testing too much reasonable stuff, but, well, at least the code gets executed
    async with app.run_test():
        with PILImage.open(TEST_IMAGE) as test_image:
            app.query_one(Image).image = ImageOps.flip(test_image)
        assert app.query_one(Image).image != TEST_IMAGE


async def test_async_loading() -> None:
    image = Image()
    image._renderable = MagicMock(is_prepared=Mock(return_value=False))
    with patch.object(image, "_trigger_prepare_async") as trigger_prepare_async:
        image.render()
    assert not trigger_prepare_async.called

    image = Image(load_async=True)
    image._renderable = MagicMock(is_prepared=Mock(return_value=False))
    with patch.object(image, "_trigger_prepare_async") as trigger_prepare_async:
        image.render()
    assert trigger_prepare_async.called


@pytest.mark.filterwarnings("ignore:coroutine 'Widget._watch_loading' was never awaited")
async def test_trigger_prepare_async() -> None:
    class TestApp(App[None]):
        def compose(self) -> ComposeResult:
            yield Image(TEST_IMAGE, load_async=True)

    app = TestApp()

    async with app.run_test():
        image = app.query_one(Image)
        with patch.object(image._renderable, "prepare_async") as prepare_async:
            await image._trigger_prepare_async(SIZE_ZERO, TERM_SIZE).wait()
        assert prepare_async.called


def test_render_no_renderable() -> None:
    image = Image()

    with patch("textual_kitty.widget.get_terminal_size_info") as get_terminal_size_info:
        image.render()
    assert not get_terminal_size_info.called
